/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gui.h"
#if GTK_CHECK_VERSION(2,12,0)
#include <gtkhtml/gtkhtml.h>
#else
#include <glade/glade.h>
#endif
#include "donate.xpm"
#include "icon_64x64.xpm"

#ifdef WIN32
#include <windows.h>
#endif

#if GTK_CHECK_VERSION(2,12,0)
static GtkBuilder *builder = NULL;
#else
static GladeXML *xml = NULL;
#endif
NNTPGrabCore *core = NULL;

void debug_initialize(void);
void info_initialize(void);
void config_initialize(void);
void queue_initialize(void);
void callbacks_initialize(void);
void nzb_initialize(void);
void connections_initialize(void);
void notify_initialize(void);
void monitor_initialize(void);
#if GTK_CHECK_VERSION(2,12,0)
void initialize_html_widget(void);
#endif

void on_btnQuit_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowMain = get_widget("windowMain");

    gtk_widget_hide(windowMain);
    gtk_main_quit();
}

#if GTK_CHECK_VERSION(2,12,0)
GObject *
get_object(const char *name)
{
    GObject *obj;

    if (!builder) {
        return NULL;
    }

    obj = gtk_builder_get_object(builder, name);
    if (!obj) {
        g_error(_("No widget found with name '%s'\n"), name);
    }

    return obj;
}

GtkWidget *
get_widget(const char *name)
{
    return GTK_WIDGET(get_object(name));
}

#else

GtkWidget *
get_widget(const char *name)
{
    GtkWidget *widget;

    if (!xml) {
        return NULL;
    }

    widget = glade_xml_get_widget(xml, name);
    if (!widget) {
        g_error(_("No widget found with name '%s'\n"), name);
    }

    return widget;
}
#endif

static void
log_func (const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer user_data)
{
    char *msg;
    GtkWidget *window;

    window = get_widget("windowMain");
    if (log_domain) {
        msg = g_strdup_printf("%s: %s", log_domain, message);
    } else {
        msg = g_strdup(message);
    }
    dialog_show(window, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
    g_free(msg);
}

static gboolean
initialisation(gpointer data)
{
    char *errmsg = NULL;
    char *warnings = NULL;

    if (!(core = nntpgrab_core_init(NNTPGRAB_API_VERSION, &errmsg, &warnings))) {
        g_error(_("NNTPGrab Initialisation FAILED:\n%s\n"), errmsg);
        g_free(errmsg);
        return 1;
    }

    if (warnings) {
        GtkWidget *windowMain = get_widget("windowMain");
        dialog_show(windowMain, warnings, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
        g_free(warnings);
    }

    debug_initialize();
    callbacks_initialize();
    info_initialize();
    config_initialize();
    nzb_initialize();
    connections_initialize();
    queue_initialize();
    notify_initialize();
    monitor_initialize();

    return FALSE;
}

int
main(int argc, char *argv[])
{
    char *filename;
    GError *err = NULL;
    GtkWidget *windowMain;
    GtkWidget *btnDonate;
    GtkWidget *icon;
    GdkPixbuf *pixbuf;
    const char *path;
#if GTK_CHECK_VERSION(2,12,0)
    const char *interface_file = "nntpgrab_gui.ui";
#else
    const char *interface_file = "nntpgrab_gui.glade";
#endif

#ifdef WIN32
    // exchndl.dll is part of Dr. MinGW. By loading this DLL, every time the
    // program crashes a .RPT file will be created containing the backtrace
    LoadLibrary("exchndl.dll");
#endif

    g_set_application_name("NNTPGrab");
    g_set_prgname("NNTPGrab");
    gtk_init(&argc, &argv);

    bindtextdomain (GETTEXT_PACKAGE, NNTPGRABLOCALEDIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    g_log_set_handler(NULL, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);
    g_log_set_handler("NNTPGrab", G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_func, NULL);

    /* load the interface */
#ifdef WIN32
    path = "..\\share\\nntpgrab";
#else
    if (g_getenv("NNTPGRAB_DATA_DIR")) {
        path = g_getenv("NNTPGRAB_DATA_DIR");
    } else {
        path = SHARE_DIR;
    }
#endif

    filename = g_build_filename(path, G_DIR_SEPARATOR_S, interface_file, NULL);

#if GTK_CHECK_VERSION(2,12,0)
    // Register GtkHTML with GObject so that GtkBuilder can use it
    gtk_html_get_type();

    builder = gtk_builder_new();
    if (!gtk_builder_add_from_file(builder, filename, &err)) {
        builder = NULL;
        g_error(err->message);
    }

    gtk_builder_connect_signals(builder, NULL);

    // GtkBuilder doesn't support a custom creation function yet, so the GtkHTML
    // component needs to be manually configured
    initialize_html_widget();
#else
    xml = glade_xml_new(filename, NULL, NULL);

    /* connect the signals in the interface */
    glade_xml_signal_autoconnect(xml);
#endif

    g_free(filename);

    // Set the icon of the window
    windowMain = get_widget("windowMain");
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) &icon_64x64_xpm);
    gtk_window_set_icon(GTK_WINDOW(windowMain), pixbuf);
    g_object_unref(pixbuf);

    // Set the donate icon of the window
    btnDonate = get_widget("btnDonate");
    pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) &donate_xpm);
    icon = gtk_image_new_from_pixbuf(pixbuf);
    gtk_widget_show(icon);
    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON(btnDonate), icon);
    g_object_unref(pixbuf);

    // Perform the main initialisation
    g_idle_add(initialisation, NULL);

    /* start the event loop */
    gtk_main();

    nntpgrab_core_cleanup(core);

    return 0;
}

