/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <glib-object.h>

#include "nntpgrab_internal.h"
#include "decoder.h"
#include "plugin_decoder.h"
#include "nntpgrab_plugin_decoder.h"

static PluginDecoderExportedFuncs plugin_funcs;

typedef struct DecoderClass DecoderClass;

static GType decoder_get_type (void);

struct Decoder
{
    GObject parent;

    GStaticRWLock rwlock;
    Configuration *config;
};

struct DecoderClass
{
    GObjectClass parent;
};

G_DEFINE_TYPE(Decoder, decoder, G_TYPE_OBJECT);

static void
decoder_init (Decoder *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
}

static void
decoder_finalize (GObject *obj)
{
    Decoder *decoder = DECODER(obj);

    g_static_rw_lock_writer_lock(&decoder->rwlock);

    g_object_unref(decoder->config);

    g_static_rw_lock_writer_unlock(&decoder->rwlock);
    g_static_rw_lock_free(&decoder->rwlock);
}

static void
decoder_class_init (DecoderClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = decoder_finalize;
}

Decoder *
decoder_new(Configuration *config)
{
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    static GModule *plugin = NULL;
    Decoder *decoder;

    g_static_mutex_lock(&mutex);
    if (!plugin) {
        PluginDecoderImportedFuncs imported_funcs;
        plugin_decoder_init_func init_func;
        char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#ifdef DARWIN
#define EXT ".dylib"
#else
#define EXT ".so"
#endif
#endif

       if (g_getenv("NNTPGRAB_LIBDIR")) {
            filename = g_strdup_printf("%s/libnntpgrab_plugin_decoder%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
       } else {
            filename = g_strdup_printf("libnntpgrab_plugin_decoder%s", EXT);
       }

        plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);

        g_free(filename);

        if (!plugin) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

#if 0
        imported_funcs.update_progress = update_progress_callback;
        imported_funcs.notify_download_complete = download_complete_callback;
        imported_funcs.notify_part_not_available = part_not_available_callback;
        imported_funcs.notify_connecting = connecting_callback;
        imported_funcs.notify_connected = connected_callback;
        imported_funcs.notify_disconnect = disconnect_callback;
#endif

        if (!g_module_symbol(plugin, "nntpgrab_plugin_decoder_get_version", (gpointer *) &plugin_funcs.get_version)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
            char *errmsg = g_strdup_printf(_("Decoder Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
            nntpgrab_core_emit_fatal_error(errmsg);
            g_free(errmsg);
            return NULL;
        }

        /* All required methods are found, lets start */
        if (!g_module_symbol(plugin, "nntpgrab_plugin_decoder_initialize", (gpointer *) &init_func)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_decoder_decode_file", (gpointer *) &plugin_funcs.decode_file)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!init_func(imported_funcs)) {
            nntpgrab_core_emit_fatal_error(_("Decoder Plugin initialization failed!"));
            return NULL;
        }

        nntpgrab_core_emit_debug_message(_("Decoder plugin initialized"));
    }

    decoder = g_object_new(DECODER_TYPE_OBJECT, NULL);
    decoder->config = config;
    g_object_ref(config);

    g_static_mutex_unlock(&mutex);

    return decoder;
}

void
decoder_destroy(Decoder *obj)
{
    g_object_unref(obj);
}

gboolean
decoder_decode_file(Decoder *obj, const char *collection_name, const NNTPFile *file, int *saved_errno)
{
    Decoder *decoder = DECODER(obj);
    ConfigOpts opts;
    char *msg;
    gboolean retval;
    char *errmsg = NULL;

    msg = g_strdup_printf(_("Trying to decode file with from collection '%s' with subject = '%s'"), collection_name, file->subject);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    g_static_rw_lock_reader_lock(&decoder->rwlock);

    opts = configuration_get_opts(decoder->config);

    g_static_rw_lock_reader_unlock(&decoder->rwlock);

    retval = plugin_funcs.decode_file(collection_name, file, opts.temp_directory, opts.download_directory, saved_errno, &errmsg);

    if (!retval && errmsg) {
        char *filename = NULL;

        nntpgrab_utils_strip_subject(file->subject, NULL, NULL, NULL, &filename, NULL, NULL, NULL, NULL);
        msg = g_strdup_printf(_("Decode of file '%s' failed:\n%s"), filename, errmsg);
        nntpgrab_core_emit_warning(msg);

        g_free(filename);
        g_free(errmsg);
    }

    msg = g_strdup_printf(_("Decoder returned %i"), retval);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    return retval;
}
