/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmodule.h>
#include <glib-object.h>

#include "configuration.h"
#include "schedular.h"
#include "plugin_schedular.h"
#include "nntpgrab_plugin_schedular.h"
#include "nntpconnection.h"
#include "decoder.h"
#include "nntpgrab_internal.h"

static PluginSchedularExportedFuncs plugin_funcs;
static GModule *plugin = NULL;
static Schedular *schedular = NULL;

typedef struct SchedularClass SchedularClass;

static GType schedular_get_type (void);

static void task_added_callback(const char *collection_name, const NNTPFile *file);
static void task_removed_callback(const char *collection_name, const char *subject);
static void task_restarted_callback(const char *collection_name, const char *subject);
static void collection_added_callback(const char *collection_name);
static void collection_removed_callback(const char *collection_name);
static GList *config_get_avail_servers(void);
static ConfigServer *config_get_server_info(const char *servername);
static gpointer nntp_connection_initialize (const char *servername, NNTPGrabErrCode *err);
static void nntp_connection_destroy (gpointer obj);
static NNTPGrabErrCode nntp_connection_get_part (gpointer obj, const char *collection_name, const NNTPFile *file, const NNTPPart *part, char **reason);
static gboolean decode_file (const char *collection_name, const NNTPFile *file, int *saved_errno);
static void notify_download_done (const char *collection_name, const char *subject);
static void notify_decode_begin (const char *collection_name, const char *subject);
static void notify_decode_done (const char *collection_name, const char *subject);
static void notify_schedular_state_changed (SchedularState new_state, const char *reason);
static void emit_fatal_error (const char *errmsg);
static void config_changed_callback(Configuration *config, ConfigChangedFlag flag, gpointer data);

struct Schedular
{
    GObject parent;

    GStaticRWLock rwlock;
    Configuration *config;
    ConfigOpts opts;
    Decoder *decoder;
};

struct SchedularClass
{
    GObjectClass parent;
};

G_DEFINE_TYPE(Schedular, schedular, G_TYPE_OBJECT);

static void
schedular_init (Schedular *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
}

static void
schedular_finalize (GObject *obj)
{
    Schedular *schedular = SCHEDULAR(obj);

    if (plugin_funcs.destroy) {
        plugin_funcs.destroy();
    }

    g_module_close(plugin);

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    g_object_unref(schedular->config);
    decoder_destroy(schedular->decoder);

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    g_static_rw_lock_free(&schedular->rwlock);
}

static void
schedular_class_init (SchedularClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = schedular_finalize;
}

Schedular *
schedular_new(Configuration *config)
{
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

    g_static_mutex_lock(&mutex);

    g_assert(schedular == NULL);

    if (!plugin) {
        char *errmsg;
        PluginSchedularImportedFuncs imported_funcs;
        plugin_schedular_init_func init_func;
        char *filename;

#ifdef WIN32
#define EXT "-0.dll"
#else
#ifdef DARWIN
#define EXT ".dylib"
#else
#define EXT ".so"
#endif
#endif

       if (g_getenv("NNTPGRAB_LIBDIR")) {
            filename = g_strdup_printf("%s/libnntpgrab_plugin_schedular%s", g_getenv("NNTPGRAB_LIBDIR"), EXT);
       } else {
            filename = g_strdup_printf("libnntpgrab_plugin_schedular%s", EXT);
       }

        plugin = g_module_open(filename, G_MODULE_BIND_LOCAL);

        g_free(filename);

        if (!plugin) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        imported_funcs.task_added = task_added_callback;
        imported_funcs.task_removed = task_removed_callback;
        imported_funcs.task_restarted = task_restarted_callback;
        imported_funcs.collection_added = collection_added_callback;
        imported_funcs.collection_removed = collection_removed_callback;
        imported_funcs.config_get_avail_servers = config_get_avail_servers;
        imported_funcs.config_get_server_info = config_get_server_info;
        imported_funcs.nntp_connection_initialize = nntp_connection_initialize;
        imported_funcs.nntp_connection_destroy = nntp_connection_destroy;
        imported_funcs.nntp_connection_get_part = nntp_connection_get_part;
        imported_funcs.decode_file = decode_file;
        imported_funcs.notify_download_done = notify_download_done;
        imported_funcs.notify_decode_begin = notify_decode_begin;
        imported_funcs.notify_decode_done = notify_decode_done;
        imported_funcs.notify_schedular_state_changed = notify_schedular_state_changed;
        imported_funcs.emit_fatal_error = emit_fatal_error;

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_get_version", (gpointer *) &plugin_funcs.get_version)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        // Check if the API version matches
        if (plugin_funcs.get_version() != NNTPGRAB_PLUGIN_API_VERSION) {
            char *errmsg = g_strdup_printf(_("Schedular Plugin API mismatch (Plugin API version = %i, expected = %i)"), plugin_funcs.get_version(), NNTPGRAB_PLUGIN_API_VERSION);
            nntpgrab_core_emit_fatal_error(errmsg);
            g_free(errmsg);
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_initialize", (gpointer *) &init_func)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_destroy", (gpointer *) &plugin_funcs.destroy)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_start", (gpointer *) &plugin_funcs.start)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_stop", (gpointer *) &plugin_funcs.stop)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_get_state", (gpointer *) &plugin_funcs.get_state)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_add_task_to_queue", (gpointer *) &plugin_funcs.add_task_to_queue)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_del_task_from_queue", (gpointer *) &plugin_funcs.del_task_from_queue)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_restart_task", (gpointer *) &plugin_funcs.restart_task)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_save_queue", (gpointer *) &plugin_funcs.save_queue)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        if (!g_module_symbol(plugin, "nntpgrab_plugin_schedular_foreach_task", (gpointer *) &plugin_funcs.foreach_task)) {
            nntpgrab_core_emit_fatal_error(g_module_error());
            return NULL;
        }

        /* All required methods are found, lets start */
        if (!init_func(imported_funcs, &errmsg)) {
            char *msg = g_strdup_printf(_("Schedular Plugin initialization failed!\n%s"), errmsg);

            nntpgrab_core_emit_fatal_error(msg);

            g_free(msg);
            g_free(errmsg);

            return NULL;
        }

        nntpgrab_core_emit_debug_message(_("Schedular plugin initialized"));
    }

    g_object_ref(config);

    schedular = g_object_new(SCHEDULAR_TYPE_OBJECT, NULL);
    schedular->config = config;
    schedular->decoder = decoder_new(config);
    schedular->opts = configuration_get_opts(schedular->config);

    g_static_mutex_unlock(&mutex);

    g_signal_connect(config, "config_changed", G_CALLBACK(config_changed_callback), NULL);

    return schedular;
}

void
schedular_destroy(Schedular *obj)
{
    g_object_unref(obj);
}

gboolean
schedular_start(Schedular *obj)
{
    g_assert(plugin_funcs.start);

    return plugin_funcs.start();
}

gboolean
schedular_stop(Schedular *obj)
{
    g_assert(plugin_funcs.stop);

    return plugin_funcs.stop(NULL);
}

SchedularState
schedular_get_state (Schedular *obj)
{
    g_assert(plugin_funcs.get_state);

    return plugin_funcs.get_state();
}

gboolean
schedular_add_task_to_queue(Schedular *obj, const char *collection_name, const NNTPFile *file, char **errmsg)
{
    char *msg;

    g_assert(plugin_funcs.add_task_to_queue);

    msg = g_strdup_printf(_("New task added to queue, collection_name = '%s', subject = '%s'"), collection_name, file->subject);
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    return plugin_funcs.add_task_to_queue(collection_name, file, errmsg);
}

gboolean
schedular_del_task_from_queue(Schedular *obj, const char *collection_name, const char *subject, char **errmsg)
{
    char *msg;

    g_assert(plugin_funcs.del_task_from_queue);

    if (!subject) {
        msg = g_strdup_printf(_("Entire collection removed from queue, collection_name = '%s'"), collection_name);
    } else {
        msg = g_strdup_printf(_("Task removed from queue, collection_name = '%s', subject = '%s'"), collection_name, subject);
    }
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    return plugin_funcs.del_task_from_queue(collection_name, subject, errmsg);
}

gboolean
schedular_restart_task(Schedular *obj, const char *collection_name, const char *subject, char **errmsg)
{
    char *msg;

    g_assert(plugin_funcs.restart_task);

    if (!subject) {
        msg = g_strdup_printf(_("Entire collection restarted, collection_name = '%s'"), collection_name);
    } else {
        msg = g_strdup_printf(_("Task restarted, collection_name = '%s', subject = '%s'"), collection_name, subject);
    }
    nntpgrab_core_emit_debug_message(msg);
    g_free(msg);

    return plugin_funcs.restart_task(collection_name, subject, errmsg);
}

gboolean
schedular_save_queue(Schedular *obj, char **errmsg)
{
    g_assert(plugin_funcs.save_queue);

    return plugin_funcs.save_queue(errmsg);
}

struct _foreach_data
{
    ForeachCollectionFunc collection_func;
    ForeachFileFunc file_func;
    ForeachGroupFunc group_func;
    gpointer data;
};

static int
calc_num_parts_downloaded(GList *parts)
{
    GList *list;
    int parts_downloaded = 0;

    list = parts;
    while (list) {
        NNTPPart *part = (NNTPPart *) list->data;

        if (part->downloaded) {
            parts_downloaded++;
        }

        list = g_list_next(list);
    }

    return parts_downloaded;
}

static void
foreach_func(gpointer data, gpointer user_data)
{
    struct _foreach_data *foreach_data = (struct _foreach_data *) user_data;
    NNTPCollection *collection = (NNTPCollection *) data;
    GList *list, *list2;

    if (foreach_data->collection_func) {
        foreach_data->collection_func(collection->collection_name, collection->total_size, foreach_data->data);
    }

    if (!foreach_data->group_func && !foreach_data->file_func) {
        return;
    }

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        // FIXME: calc_num_parts_downloaded() and num_parts_failed need to be better calculated
        if (foreach_data->file_func) {
            foreach_data->file_func(collection->collection_name, file->subject, file->poster, file->stamp, file->file_size, file->position, g_list_length(file->parts), calc_num_parts_downloaded(file->parts), 0, foreach_data->data);
        }

        if (foreach_data->group_func) {
            list2 = file->groups;
            while (list2) {
                const char *group = (const char *) list2->data;

                foreach_data->group_func(collection->collection_name, file->subject, group, foreach_data->data);

                list2 = g_list_next(list2);
            }
        }

        list = g_list_next(list);
    }
}

void
schedular_foreach_task(Schedular *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, gpointer data)
{
    struct _foreach_data foreach_data;

    g_assert(plugin_funcs.foreach_task);

    foreach_data.collection_func = collection_func;
    foreach_data.file_func = file_func;
    foreach_data.group_func = group_func;
    foreach_data.data = data;

    plugin_funcs.foreach_task(foreach_func, &foreach_data);
}

static void
task_added_callback(const char *collection_name, const NNTPFile *file)
{
    nntpgrab_core_emit_file_added(collection_name, file);
}

static void
task_removed_callback(const char *collection_name, const char *subject)
{
    nntpgrab_core_emit_file_removed(collection_name, subject);
}

static void
task_restarted_callback(const char *collection_name, const char *subject)
{
    nntpgrab_core_emit_file_restarted(collection_name, subject);
}

static void
collection_added_callback(const char *collection_name)
{
    nntpgrab_core_emit_collection_added(collection_name);
}

static void
collection_removed_callback(const char *collection_name)
{
    nntpgrab_core_emit_collection_removed(collection_name);
}

static GList *
config_get_avail_servers(void)
{
    g_assert(schedular);

    return configuration_get_avail_servers(schedular->config);
}

static ConfigServer *
config_get_server_info(const char *servername)
{
    g_assert(schedular);

    return configuration_get_server_info(schedular->config, servername);
}

static gpointer
nntp_connection_initialize (const char *servername, NNTPGrabErrCode *err)
{
    NNTPConnection *conn;
    ConfigServer *server;

    server = configuration_get_server_info(schedular->config, servername);
    if (!server) {
        return NULL;
    }

    conn = nntpconnection_new(schedular->config, *server);
    g_free(server);

    if ((*err = nntpconnection_connect(conn)) != NNTP_ERROR_NONE) {
        nntpconnection_destroy(conn);
        return NULL;
    }

    return (gpointer) conn;
}

static void
nntp_connection_destroy (gpointer obj)
{
    g_assert(obj != NULL);

    if (nntpconnection_get_is_connected(obj)) {
        nntpconnection_disconnect(obj);
    }

    nntpconnection_destroy(obj);
}

static NNTPGrabErrCode
nntp_connection_get_part (gpointer obj, const char *collection_name, const NNTPFile *file, const NNTPPart *part, char **reason)
{
    char *filename;
    NNTPGrabErrCode ret;

    g_assert(obj != NULL);
    g_assert(schedular != NULL);

    // Is the connection still alive?
    if (!nntpconnection_get_is_connected(obj)) {
        // Nope, reconnect first
        nntpgrab_core_emit_debug_message(_("Connection lost. Trying to reconnect"));
        if ((ret = nntpconnection_connect(obj)) != NNTP_ERROR_NONE) {
            nntpgrab_core_emit_debug_message(_("Reconnect failed"));
            nntpconnection_disconnect(obj);
            return ret;
        }
    }

    g_static_rw_lock_reader_lock(&schedular->rwlock);
    filename = g_strdup_printf("%s%s%s.%i", schedular->opts.temp_directory, G_DIR_SEPARATOR_S, file->tmp_filename, part->partnum);
    g_static_rw_lock_reader_unlock(&schedular->rwlock);

    if ((ret = nntpconnection_get_part(obj, collection_name, file->subject, part, filename)) != NNTP_ERROR_NONE) {
        g_free(filename);
        return ret;
    }

    g_free(filename);

    nntpgrab_core_emit_debug_message(_("Saving of part SUCCEEDED"));

    return NNTP_ERROR_NONE;
}

static gboolean
decode_file (const char *collection_name, const NNTPFile *file, int *saved_errno)
{
    return decoder_decode_file(schedular->decoder, collection_name, file, saved_errno);
}

static void
notify_download_done (const char *collection_name, const char *subject)
{
    nntpgrab_core_emit_file_download_done(collection_name, subject);
}

static void
notify_decode_begin (const char *collection_name, const char *subject)
{
    nntpgrab_core_emit_file_decode_begin(collection_name, subject);
}

static void
notify_decode_done (const char *collection_name, const char *subject)
{
    nntpgrab_core_emit_file_decode_done(collection_name, subject);
}

static void
notify_schedular_state_changed (SchedularState new_state, const char *reason)
{
    nntpgrab_core_emit_schedular_state_changed(new_state, reason);
}

static void
emit_fatal_error (const char *errmsg)
{
    nntpgrab_core_emit_fatal_error(errmsg);
}

static void
config_changed_callback (Configuration *config, ConfigChangedFlag flag, gpointer data)
{
    g_assert(schedular);

    nntpgrab_core_emit_debug_message(_("Schedular: Received config changed signal"));

    if (flag == CONFIG_CHANGED_OPTS_CHANGED) {
        g_static_rw_lock_writer_lock(&schedular->rwlock);
        schedular->opts = configuration_get_opts(config);
        g_static_rw_lock_writer_unlock(&schedular->rwlock);
    }

    if (flag == CONFIG_CHANGED_SERVER_ADDED || flag == CONFIG_CHANGED_SERVER_DELETED) {
        if (schedular_get_state(schedular) == SCHEDULAR_STATE_RUNNING) {
            // Restart the schedular to apply the new settings
            schedular_stop(schedular);
            schedular_start(schedular);
        }
    }
}
